@@ -7,8 +7,9 @@
 *     Operator of Los Alamos National Laboratory.
 * Copyright (c) 2002 Berliner Elektronenspeicherringgesellschaft fuer
 *     Synchrotronstrahlung.
+* SPDX-License-Identifier: EPICS
 * EPICS BASE is distributed subject to a Software License Agreement found
-* in file LICENSE that is included with this distribution. 
+* in file LICENSE that is included with this distribution.
 \*************************************************************************/
 
 /*
@@ -109,12 +110,12 @@ static void usage (void)
 
 /*+**************************************************************************
  *
- * Function:	event_handler
+ * Function:    event_handler
  *
- * Description:	CA event_handler for request type callback
- * 		Allocates the dbr structure and copies the data
+ * Description: CA event_handler for request type callback
+ *              Allocates the dbr structure and copies the data
  *
- * Arg(s) In:	args  -  event handler args (see CA manual)
+ * Arg(s) In:   args  -  event handler args (see CA manual)
  *
  **************************************************************************-*/
 
@@ -137,22 +138,22 @@ static void event_handler (evargs args)
 
 /*+**************************************************************************
  *
- * Function:	caget
+ * Function:    caget
  *
- * Description:	Issue read requests, wait for incoming data
- * 		and print the data according to the selected format
+ * Description: Issue read requests, wait for incoming data
+ *              and print the data according to the selected format
  *
- * Arg(s) In:	pvs       -  Pointer to an array of pv structures
+ * Arg(s) In:   pvs       -  Pointer to an array of pv structures
  *              nPvs      -  Number of elements in the pvs array
  *              request   -  Request type
  *              format    -  Output format
  *              dbrType   -  Requested dbr type
  *              reqElems  -  Requested number of (array) elements
  *
- * Return(s):	Error code: 0 = OK, 1 = Error
+ * Return(s):   Error code: 0 = OK, 1 = Error
  *
  **************************************************************************-*/
- 
+
 static int caget (pv *pvs, int nPvs, RequestT request, OutputT format,
            chtype dbrType, unsigned long reqElems)
 {
@@ -255,7 +256,7 @@ static int caget (pv *pvs, int nPvs, RequestT request, OutputT format,
     for (n = 0; n < nPvs; n++) {
 
         switch (format) {
-        case plain:             /* Emulate old caget behaviour */
+        case plain:             /* Emulate old caget behavior */
             if (pvs[n].nElems <= 1 && fieldSeparator == ' ') printf("%-30s", pvs[n].name);
             else                                               printf("%s", pvs[n].name);
             printf("%c", fieldSeparator);
@@ -342,6 +343,7 @@ static int caget (pv *pvs, int nPvs, RequestT request, OutputT format,
         default :
             break;
         }
+        free(pvs[n].value);
     }
     return 0;
 }
@@ -350,23 +352,23 @@ static int caget (pv *pvs, int nPvs, RequestT request, OutputT format,
 
 /*+**************************************************************************
  *
- * Function:	main
+ * Function:    main
  *
- * Description:	caget main()
- * 		Evaluate command line options, set up CA, connect the
- * 		channels, collect and print the data as requested
+ * Description: caget main()
+ *              Evaluate command line options, set up CA, connect the
+ *              channels, collect and print the data as requested
  *
- * Arg(s) In:	[options] <pv-name> ...
+ * Arg(s) In:   [options] <pv-name> ...
  *
- * Arg(s) Out:	none
+ * Arg(s) Out:  none
  *
- * Return(s):	Standard return code (0=success, 1=error)
+ * Return(s):   Standard return code (0=success, 1=error)
  *
  **************************************************************************-*/
 
 static void complainIfNotPlainAndSet (OutputT *current, const OutputT requested)
 {
-    if (*current != plain) 
+    if (*current != plain)
         fprintf(stderr,
                 "Options t,d,a are mutually exclusive. "
                 "('caget -h' for help.)\n");
@@ -387,10 +389,12 @@ int main (int argc, char *argv[])
     int digits = 0;             /* getopt() no. of float digits */
 
     int nPvs;                   /* Number of PVs */
-    pv* pvs;                    /* Array of PV structures */
+    pv* pvs = NULL;             /* Array of PV structures */
 
     LINE_BUFFER(stdout);        /* Configure stdout buffering */
 
+    use_ca_timeout_env ( &caTimeout);
+
     while ((opt = getopt(argc, argv, ":taicnhsSVe:f:g:l:#:d:0:w:p:F:")) != -1) {
         switch (opt) {
         case 'h':               /* Print usage */
@@ -436,8 +440,8 @@ int main (int argc, char *argv[])
             if(epicsScanDouble(optarg, &caTimeout) != 1)
             {
                 fprintf(stderr, "'%s' is not a valid timeout value "
-                        "- ignored. ('caget -h' for help.)\n", optarg);
-                caTimeout = DEFAULT_TIMEOUT;
+                        "- ignored, using '%.1f'. ('caget -h' for help.)\n",
+                        optarg, caTimeout);
             }
             break;
         case '#':               /* Array count */
@@ -467,7 +471,7 @@ int main (int argc, char *argv[])
         case 'f':
         case 'g':
             if (sscanf(optarg, "%d", &digits) != 1)
-                fprintf(stderr, 
+                fprintf(stderr,
                         "Invalid precision argument '%s' "
                         "for option '-%c' - ignored.\n", optarg, opt);
             else
@@ -553,6 +557,7 @@ int main (int argc, char *argv[])
         result = caget(pvs, nPvs, request, format, type, count);
 
                                 /* Shut down Channel Access */
+    free(pvs);
     ca_context_destroy();
 
     return result;
