@@ -6,6 +6,7 @@
 *     National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 *     Operator of Los Alamos National Laboratory.
+* SPDX-License-Identifier: EPICS
 * EPICS BASE is distributed subject to a Software License Agreement found
 * in file LICENSE that is included with this distribution.
 \*************************************************************************/
@@ -34,7 +35,6 @@
 #include "errlog.h"
 #include "errMdef.h"
 
-#include "epicsExport.h" /* #define epicsExportSharedSymbols */
 #include "caeventmask.h"
 #include "callback.h"
 #include "dbAccessDefs.h"
@@ -61,18 +61,19 @@
 #include "recGbl.h"
 #include "recSup.h"
 #include "special.h"
+#include "epicsExport.h"
 
-epicsShareDef struct dbBase *pdbbase = 0;
-epicsShareDef volatile int interruptAccept=FALSE;
+struct dbBase *pdbbase = 0;
+volatile int interruptAccept=FALSE;
 
-epicsShareDef int dbAccessDebugPUTF = 0;
+int dbAccessDebugPUTF = 0;
 epicsExportAddress(int, dbAccessDebugPUTF);
 
 /* Hook Routines */
 
-epicsShareDef DB_LOAD_RECORDS_HOOK_ROUTINE dbLoadRecordsHook = NULL;
+DB_LOAD_RECORDS_HOOK_ROUTINE dbLoadRecordsHook = NULL;
 
-static short mapDBFToDBR[DBF_NTYPES] = {
+static const short mapDBFToDBR[DBF_NTYPES] = {
     /* DBF_STRING   => */    DBR_STRING,
     /* DBF_CHAR     => */    DBR_CHAR,
     /* DBF_UCHAR    => */    DBR_UCHAR,
@@ -111,176 +112,179 @@ void dbSpcAsRegisterCallback(SPC_ASCALLBACK func)
 
 long dbPutSpecial(DBADDR *paddr,int pass)
 {
-    long int	(*pspecial)()=NULL;
+    long int    (*pspecial)(struct dbAddr *, int)=NULL;
     rset        *prset;
-    dbCommon 	*precord = paddr->precord;
-    long	status=0;
-    long	special=paddr->special;
+    dbCommon    *precord = paddr->precord;
+    long        status=0;
+    long        special=paddr->special;
 
     prset = dbGetRset(paddr);
     if(special<100) { /*global processing*/
-	if((special==SPC_NOMOD) && (pass==0)) {
-	    status = S_db_noMod;
-	    recGblDbaddrError(status,paddr,"dbPut");
-	    return(status);
-	}else if(special==SPC_SCAN){
-	    if(pass==0)
-		scanDelete(precord);
-	    else
-		scanAdd(precord);
-	}else if((special==SPC_AS) && (pass==1)) {
+        if((special==SPC_NOMOD) && (pass==0)) {
+            status = S_db_noMod;
+            recGblDbaddrError(status,paddr,"dbPut");
+            return(status);
+        }else if(special==SPC_SCAN){
+            if(pass==0)
+                scanDelete(precord);
+            else
+                scanAdd(precord);
+        }else if((special==SPC_AS) && (pass==1)) {
             if(spcAsCallback) (*spcAsCallback)(precord);
-	}
+        }
     }else {
-	if( prset && (pspecial = (prset->special))) {
-	    status=(*pspecial)(paddr,pass);
-	    if(status) return(status);
-	} else if(pass==0){
-	    recGblRecSupError(S_db_noSupport,paddr,"dbPut", "special");
-	    return(S_db_noSupport);
-	}
+        if( prset && (pspecial = (prset->special))) {
+            status=(*pspecial)(paddr,pass);
+            if(status) return(status);
+        } else if(pass==0){
+            recGblRecSupError(S_db_noSupport,paddr,"dbPut", "special");
+            return(S_db_noSupport);
+        }
     }
     return(0);
 }
 
 static void get_enum_strs(DBADDR *paddr, char **ppbuffer,
-    rset *prset,long	*options)
+                          rset *prset, long *options)
 {
-	short		field_type=paddr->field_type;
-	dbFldDes	*pdbFldDes = paddr->pfldDes;
-	dbMenu		*pdbMenu;
-	dbDeviceMenu	*pdbDeviceMenu;
-	char		**papChoice;
-	unsigned long	no_str;
-	char		*ptemp;
-	struct dbr_enumStrs *pdbr_enumStrs=(struct dbr_enumStrs*)(*ppbuffer);
-    unsigned int i;
+    struct dbr_enumStrs *penum=(struct dbr_enumStrs*)(*ppbuffer);
+    /* advance output buffer on success or failure for next option */
+    *ppbuffer = dbr_enumStrs_size + (char*)penum;
 
-	memset(pdbr_enumStrs,'\0',dbr_enumStrs_size);
-	switch(field_type) {
-		case DBF_ENUM:
-		    if( prset && prset->get_enum_strs ) {
-			(*prset->get_enum_strs)(paddr,pdbr_enumStrs);
-		    } else {
-			*options = (*options)^DBR_ENUM_STRS;/*Turn off option*/
-		    }
-		    break;
-		case DBF_MENU:
-		    pdbMenu = (dbMenu *)pdbFldDes->ftPvt;
-		    no_str = pdbMenu->nChoice;
-		    papChoice= pdbMenu->papChoiceValue;
-		    goto choice_common;
-		case DBF_DEVICE:
-		    pdbDeviceMenu = (dbDeviceMenu *)pdbFldDes->ftPvt;
-                    if(!pdbDeviceMenu) {
-                        *options = (*options)^DBR_ENUM_STRS;/*Turn off option*/
-                        break;
-                    }
-		    no_str = pdbDeviceMenu->nChoice;
-		    papChoice = pdbDeviceMenu->papChoice;
-		    goto choice_common;
-choice_common:
-		    i = sizeof(pdbr_enumStrs->strs)/
-			sizeof(pdbr_enumStrs->strs[0]);
-		    if(i<no_str) no_str = i;
-		    pdbr_enumStrs->no_str = no_str;
-		    ptemp = &(pdbr_enumStrs->strs[0][0]);
-		    for (i=0; i<no_str; i++) {
-			if(papChoice[i]==NULL) *ptemp=0;
-			else {
-			    strncpy(ptemp,papChoice[i],
-				sizeof(pdbr_enumStrs->strs[0]));
-			    *(ptemp+sizeof(pdbr_enumStrs->strs[0])-1) = 0;
-			}
-			ptemp += sizeof(pdbr_enumStrs->strs[0]);
-		    }
-		    break;
-		default:
-		    *options = (*options)^DBR_ENUM_STRS;/*Turn off option*/
-		    break;
-	}
-	*ppbuffer = ((char *)*ppbuffer) + dbr_enumStrs_size;
-	return;
+    memset(penum, 0, dbr_enumStrs_size);
+
+    /* from this point
+     * on success, return early
+     * on failure, jump or fall through to clear *options bit.
+     */
+
+    if(paddr->field_type == DBF_ENUM) {
+        if( prset && prset->get_enum_strs ) {
+            (*prset->get_enum_strs)(paddr,penum);
+            return;
+        }
+
+    } else if(paddr->field_type == DBF_MENU || paddr->field_type == DBF_DEVICE) {
+        char **ppchoices;
+        epicsUInt32 i, nchoices = 0;
+
+        if(paddr->field_type == DBF_MENU) {
+            dbMenu *pmenu = paddr->pfldDes->ftPvt;
+            nchoices = pmenu->nChoice;
+            ppchoices= pmenu->papChoiceValue;
+
+        } else if(paddr->field_type == DBF_DEVICE) {
+            dbDeviceMenu *pdevs = paddr->pfldDes->ftPvt;
+            if(!pdevs)
+                goto nostrs;
+
+            nchoices = pdevs->nChoice;
+            ppchoices = pdevs->papChoice;
+        }
+
+        if(nchoices > NELEMENTS(penum->strs))
+            nchoices = NELEMENTS(penum->strs); /* availible > capacity, truncated list */
+
+        penum->no_str = nchoices;
+
+        for(i=0; i<nchoices; i++) {
+            if(ppchoices[i]) {
+                strncpy(penum->strs[i], ppchoices[i],
+                        sizeof(penum->strs[i]));
+                /* strs[i][] allowed to omit trailing nil */
+            } else {
+                penum->strs[i][0] = '\0';
+            }
+        }
+        return;
+
+    } else {
+        /* other DBF_* fall through to error */
+    }
+
+nostrs:
+    /* indicate option data not available.  distinct from no_str==0 */
+    *options = (*options)^DBR_ENUM_STRS;
 }
 
 static void get_graphics(DBADDR *paddr, char **ppbuffer,
-    rset *prset,long	*options)
+    rset *prset,long    *options)
 {
-	struct			dbr_grDouble grd;
-	int			got_data=FALSE;
+        struct          dbr_grDouble grd;
+        int             got_data=FALSE;
 
         grd.upper_disp_limit = grd.lower_disp_limit = 0.0;
-	if( prset && prset->get_graphic_double ) {
-		(*prset->get_graphic_double)(paddr,&grd);
-		got_data=TRUE;
-	}
-	if( (*options) & (DBR_GR_LONG) ) {
-		char	*pbuffer=*ppbuffer;
+        if( prset && prset->get_graphic_double ) {
+                (*prset->get_graphic_double)(paddr,&grd);
+                got_data=TRUE;
+        }
+        if( (*options) & (DBR_GR_LONG) ) {
+                char    *pbuffer=*ppbuffer;
 
-		if(got_data) {
-		    struct dbr_grLong *pgr=(struct dbr_grLong*)pbuffer;
-		    pgr->upper_disp_limit = (epicsInt32)grd.upper_disp_limit;
-		    pgr->lower_disp_limit = (epicsInt32)grd.lower_disp_limit;
-		} else {
-		    memset(pbuffer,'\0',dbr_grLong_size);
-		    *options = (*options) ^ DBR_GR_LONG; /*Turn off option*/
-		}
-		*ppbuffer = ((char *)*ppbuffer) + dbr_grLong_size;
-	}
-	if( (*options) & (DBR_GR_DOUBLE) ) {
-		char	*pbuffer=*ppbuffer;
+                if(got_data) {
+                    struct dbr_grLong *pgr=(struct dbr_grLong*)pbuffer;
+                    pgr->upper_disp_limit = (epicsInt32)grd.upper_disp_limit;
+                    pgr->lower_disp_limit = (epicsInt32)grd.lower_disp_limit;
+                } else {
+                    memset(pbuffer,'\0',dbr_grLong_size);
+                    *options = (*options) ^ DBR_GR_LONG; /*Turn off option*/
+                }
+                *ppbuffer = ((char *)*ppbuffer) + dbr_grLong_size;
+        }
+        if( (*options) & (DBR_GR_DOUBLE) ) {
+                char    *pbuffer=*ppbuffer;
 
-		if(got_data) {
-		    struct dbr_grDouble *pgr=(struct dbr_grDouble*)pbuffer;
-		    pgr->upper_disp_limit = grd.upper_disp_limit;
-		    pgr->lower_disp_limit = grd.lower_disp_limit;
-		} else {
-		    memset(pbuffer,'\0',dbr_grDouble_size);
-		    *options = (*options) ^ DBR_GR_DOUBLE; /*Turn off option*/
-		}
-		*ppbuffer = ((char *)*ppbuffer) + dbr_grDouble_size;
-	}
-	return;
+                if(got_data) {
+                    struct dbr_grDouble *pgr=(struct dbr_grDouble*)pbuffer;
+                    pgr->upper_disp_limit = grd.upper_disp_limit;
+                    pgr->lower_disp_limit = grd.lower_disp_limit;
+                } else {
+                    memset(pbuffer,'\0',dbr_grDouble_size);
+                    *options = (*options) ^ DBR_GR_DOUBLE; /*Turn off option*/
+                }
+                *ppbuffer = ((char *)*ppbuffer) + dbr_grDouble_size;
+        }
+        return;
 }
 
 static void get_control(DBADDR *paddr, char **ppbuffer,
-    rset *prset,long	*options)
+    rset *prset,long    *options)
 {
-	struct dbr_ctrlDouble	ctrld;
-	int			got_data=FALSE;
+        struct dbr_ctrlDouble   ctrld;
+        int                     got_data=FALSE;
 
         ctrld.upper_ctrl_limit = ctrld.lower_ctrl_limit = 0.0;
-	if( prset && prset->get_control_double ) {
-		(*prset->get_control_double)(paddr,&ctrld);
-		got_data=TRUE;
-	}
-	if( (*options) & (DBR_CTRL_LONG) ) {
-		char	*pbuffer=*ppbuffer;
+        if( prset && prset->get_control_double ) {
+                (*prset->get_control_double)(paddr,&ctrld);
+                got_data=TRUE;
+        }
+        if( (*options) & (DBR_CTRL_LONG) ) {
+                char    *pbuffer=*ppbuffer;
 
-		if(got_data) {
-		    struct dbr_ctrlLong *pctrl=(struct dbr_ctrlLong*)pbuffer;
-		    pctrl->upper_ctrl_limit = (epicsInt32)ctrld.upper_ctrl_limit;
-		    pctrl->lower_ctrl_limit = (epicsInt32)ctrld.lower_ctrl_limit;
-		} else {
-		    memset(pbuffer,'\0',dbr_ctrlLong_size);
-		    *options = (*options) ^ DBR_CTRL_LONG; /*Turn off option*/
-		}
-		*ppbuffer = ((char *)*ppbuffer) + dbr_ctrlLong_size;
-	}
-	if( (*options) & (DBR_CTRL_DOUBLE) ) {
-		char	*pbuffer=*ppbuffer;
+                if(got_data) {
+                    struct dbr_ctrlLong *pctrl=(struct dbr_ctrlLong*)pbuffer;
+                    pctrl->upper_ctrl_limit = (epicsInt32)ctrld.upper_ctrl_limit;
+                    pctrl->lower_ctrl_limit = (epicsInt32)ctrld.lower_ctrl_limit;
+                } else {
+                    memset(pbuffer,'\0',dbr_ctrlLong_size);
+                    *options = (*options) ^ DBR_CTRL_LONG; /*Turn off option*/
+                }
+                *ppbuffer = ((char *)*ppbuffer) + dbr_ctrlLong_size;
+        }
+        if( (*options) & (DBR_CTRL_DOUBLE) ) {
+                char    *pbuffer=*ppbuffer;
 
-		if(got_data) {
-		   struct dbr_ctrlDouble *pctrl=(struct dbr_ctrlDouble*)pbuffer;
-		   pctrl->upper_ctrl_limit = ctrld.upper_ctrl_limit;
-		   pctrl->lower_ctrl_limit = ctrld.lower_ctrl_limit;
-		} else {
-		    memset(pbuffer,'\0',dbr_ctrlDouble_size);
-		    *options = (*options) ^ DBR_CTRL_DOUBLE; /*Turn off option*/
-		}
-		*ppbuffer = ((char *)*ppbuffer) + dbr_ctrlDouble_size;
-	}
-	return;
+                if(got_data) {
+                   struct dbr_ctrlDouble *pctrl=(struct dbr_ctrlDouble*)pbuffer;
+                   pctrl->upper_ctrl_limit = ctrld.upper_ctrl_limit;
+                   pctrl->lower_ctrl_limit = ctrld.lower_ctrl_limit;
+                } else {
+                    memset(pbuffer,'\0',dbr_ctrlDouble_size);
+                    *options = (*options) ^ DBR_CTRL_DOUBLE; /*Turn off option*/
+                }
+                *ppbuffer = ((char *)*ppbuffer) + dbr_ctrlDouble_size;
+        }
+        return;
 }
 
 static void get_alarm(DBADDR *paddr, char **ppbuffer,
@@ -332,82 +336,101 @@ static void get_alarm(DBADDR *paddr, char **ppbuffer,
 static void getOptions(DBADDR *paddr, char **poriginal, long *options,
         void *pflin)
 {
-	db_field_log	*pfl= (db_field_log *)pflin;
-    rset	*prset;
-        short		field_type;
-	dbCommon	*pcommon;
-	char		*pbuffer = *poriginal;
+        db_field_log    *pfl= (db_field_log *)pflin;
+        rset            *prset;
+        short           field_type;
+        dbCommon        *pcommon;
+        char            *pbuffer = *poriginal;
 
-        if (!pfl || pfl->type == dbfl_type_rec)
+        if (!pfl)
             field_type = paddr->field_type;
         else
             field_type = pfl->field_type;
-	prset=dbGetRset(paddr);
-	/* Process options */
-	pcommon = paddr->precord;
-	if( (*options) & DBR_STATUS ) {
-	    unsigned short *pushort = (unsigned short *)pbuffer;
+        prset=dbGetRset(paddr);
+        /* Process options */
+        pcommon = paddr->precord;
+        if( (*options) & DBR_STATUS ) {
+            unsigned short *pushort = (unsigned short *)pbuffer;
 
-            if (!pfl || pfl->type == dbfl_type_rec) {
+            if (!pfl) {
                 *pushort++ = pcommon->stat;
                 *pushort++ = pcommon->sevr;
             } else {
                 *pushort++ = pfl->stat;
                 *pushort++ = pfl->sevr;
             }
-	    *pushort++ = pcommon->acks;
-	    *pushort++ = pcommon->ackt;
-	    pbuffer = (char *)pushort;
-	}
-	if( (*options) & DBR_UNITS ) {
-	    memset(pbuffer,'\0',dbr_units_size);
-	    if( prset && prset->get_units ){
-		(*prset->get_units)(paddr, pbuffer);
-		pbuffer[DB_UNITS_SIZE-1] = '\0';
-	    } else {
-		*options ^= DBR_UNITS; /*Turn off DBR_UNITS*/
-	    }
-	    pbuffer += dbr_units_size;
-	}
-	if( (*options) & DBR_PRECISION ) {
-	    memset(pbuffer, '\0', dbr_precision_size);
-	    if((field_type==DBF_FLOAT || field_type==DBF_DOUBLE)
-	    &&  prset && prset->get_precision ){
+            *pushort++ = pcommon->acks;
+            *pushort++ = pcommon->ackt;
+            pbuffer = (char *)pushort;
+        }
+        if( (*options) & DBR_AMSG ) {
+            if (!pfl) {
+                STATIC_ASSERT(sizeof(pcommon->amsg)==sizeof(pfl->amsg));
+                strncpy(pbuffer, pcommon->amsg, sizeof(pcommon->amsg)-1);
+            } else {
+                strncpy(pbuffer, pfl->amsg,sizeof(pfl->amsg)-1);
+            }
+            pbuffer[sizeof(pcommon->amsg)-1] = '\0';
+            pbuffer += sizeof(pcommon->amsg);
+        }
+        if( (*options) & DBR_UNITS ) {
+            memset(pbuffer,'\0',dbr_units_size);
+            if( prset && prset->get_units ){
+                (*prset->get_units)(paddr, pbuffer);
+                pbuffer[DB_UNITS_SIZE-1] = '\0';
+            } else {
+                *options ^= DBR_UNITS; /*Turn off DBR_UNITS*/
+            }
+            pbuffer += dbr_units_size;
+        }
+        if( (*options) & DBR_PRECISION ) {
+            memset(pbuffer, '\0', dbr_precision_size);
+            if((field_type==DBF_FLOAT || field_type==DBF_DOUBLE)
+            &&  prset && prset->get_precision ){
                 (*prset->get_precision)(paddr,(long *)pbuffer);
-	    } else {
-		*options ^= DBR_PRECISION; /*Turn off DBR_PRECISION*/
-	    }
-	    pbuffer += dbr_precision_size;
-	}
-	if( (*options) & DBR_TIME ) {
-	    epicsUInt32 *ptime = (epicsUInt32 *)pbuffer;
+            } else {
+                *options ^= DBR_PRECISION; /*Turn off DBR_PRECISION*/
+            }
+            pbuffer += dbr_precision_size;
+        }
+        if( (*options) & DBR_TIME ) {
+            epicsUInt32 *ptime = (epicsUInt32 *)pbuffer;
 
-            if (!pfl || pfl->type == dbfl_type_rec) {
+            if (!pfl) {
                 *ptime++ = pcommon->time.secPastEpoch;
                 *ptime++ = pcommon->time.nsec;
             } else {
                 *ptime++ = pfl->time.secPastEpoch;
                 *ptime++ = pfl->time.nsec;
-	    }
-	    pbuffer = (char *)ptime;
-	}
-	if( (*options) & DBR_ENUM_STRS )
-	    get_enum_strs(paddr, &pbuffer, prset, options);
-	if( (*options) & (DBR_GR_LONG|DBR_GR_DOUBLE ))
-	    get_graphics(paddr, &pbuffer, prset, options);
-	if((*options) & (DBR_CTRL_LONG | DBR_CTRL_DOUBLE ))
-	    get_control(paddr, &pbuffer, prset, options);
-	if((*options) & (DBR_AL_LONG | DBR_AL_DOUBLE ))
-	    get_alarm(paddr, &pbuffer, prset, options);
-	*poriginal = pbuffer;
+            }
+            pbuffer = (char *)ptime;
+        }
+        if( (*options) & DBR_UTAG ) {
+            epicsUInt64 *ptag = (epicsUInt64*)pbuffer;
+            if (!pfl) {
+                *ptag++ = pcommon->utag;
+            } else {
+                *ptag++ = pfl->utag;
+            }
+            pbuffer = (char *)ptag;
+        }
+        if( (*options) & DBR_ENUM_STRS )
+            get_enum_strs(paddr, &pbuffer, prset, options);
+        if( (*options) & (DBR_GR_LONG|DBR_GR_DOUBLE ))
+            get_graphics(paddr, &pbuffer, prset, options);
+        if((*options) & (DBR_CTRL_LONG | DBR_CTRL_DOUBLE ))
+            get_control(paddr, &pbuffer, prset, options);
+        if((*options) & (DBR_AL_LONG | DBR_AL_DOUBLE ))
+            get_alarm(paddr, &pbuffer, prset, options);
+        *poriginal = pbuffer;
 }
 
 rset * dbGetRset(const struct dbAddr *paddr)
 {
-	struct dbFldDes *pfldDes = paddr->pfldDes;
+        struct dbFldDes *pfldDes = paddr->pfldDes;
 
-	if(!pfldDes) return(0);
-	return(pfldDes->pdbRecordType->prset);
+        if(!pfldDes) return(0);
+        return(pfldDes->pdbRecordType->prset);
 }
 
 long dbPutAttribute(
@@ -466,7 +489,7 @@ long dbProcess(dbCommon *precord)
     char context[40] = "";
     long status = 0;
     int *ptrace;
-    int	set_trace = FALSE;
+    int set_trace = FALSE;
     dbFldDes *pdbFldDes;
     int callNotifyCompletion = FALSE;
 
@@ -509,7 +532,7 @@ long dbProcess(dbCommon *precord)
         }
     }
 
-    /* If already active dont process */
+    /* If already active don't process */
     if (precord->pact) {
         unsigned short monitor_mask;
 
@@ -522,7 +545,7 @@ long dbProcess(dbCommon *precord)
             (precord->lcnt++ < MAX_LOCK) ||
             (precord->sevr >= INVALID_ALARM)) goto all_done;
 
-        recGblSetSevr(precord, SCAN_ALARM, INVALID_ALARM);
+        recGblSetSevrMsg(precord, SCAN_ALARM, INVALID_ALARM, "Async in progress");
         monitor_mask = recGblResetAlarms(precord);
         monitor_mask |= DBE_VALUE|DBE_LOG;
         pdbFldDes = pdbRecordType->papFldDes[pdbRecordType->indvalFlddes];
@@ -736,6 +759,8 @@ long dbValueSize(short dbr_type)
         sizeof(epicsFloat64),        /* DOUBLE       */
         sizeof(epicsEnum16)};        /* ENUM         */
 
+    if(dbr_type>=NELEMENTS(size))
+        return 0;
     return(size[dbr_type]);
 }
 
@@ -776,8 +801,14 @@ int dbLoadRecords(const char* file, const char* subs)
         return -1;
     }
     status = dbReadDatabase(&pdbbase, file, 0, subs);
-    if (!status && dbLoadRecordsHook)
-        dbLoadRecordsHook(file, subs);
+    if(status==0) {
+        if(dbLoadRecordsHook)
+            dbLoadRecordsHook(file, subs);
+    } else {
+        fprintf(stderr, ERL_ERROR " failed to load '%s'\n", file);
+        if(status==-2)
+            fprintf(stderr, "    Records cannot be loaded after iocInit!\n");
+    }
     return status;
 }
 
@@ -785,15 +816,12 @@ int dbLoadRecords(const char* file, const char* subs)
 static long getLinkValue(DBADDR *paddr, short dbrType,
     char *pbuf, long *nRequest)
 {
-    dbCommon *precord = paddr->precord;
-    dbFldDes *pfldDes = paddr->pfldDes;
     /* size of pbuf storage in bytes, including space for trailing nil */
     int maxlen;
     DBENTRY dbEntry;
-    long status;
     long nReq = nRequest ? *nRequest : 1;
 
-    /* dbFindRecord() below will always succeed as we have a
+    /* below will always succeed as we have a
      * valid DBADDR, so no point to check again.
      * Request for zero elements always succeeds
      */
@@ -819,10 +847,8 @@ static long getLinkValue(DBADDR *paddr, short dbrType,
         return S_db_badDbrtype;
     }
 
-    dbInitEntry(pdbbase, &dbEntry);
-    status = dbFindRecord(&dbEntry, precord->name);
-    if (!status) status = dbFindField(&dbEntry, pfldDes->name);
-    if (!status) {
+    dbInitEntryFromAddr(paddr, &dbEntry);
+    {
         const char *rtnString = dbGetString(&dbEntry);
 
         strncpy(pbuf, rtnString, maxlen-1);
@@ -832,7 +858,7 @@ static long getLinkValue(DBADDR *paddr, short dbrType,
         if(nRequest) *nRequest = nReq;
     }
     dbFinishEntry(&dbEntry);
-    return status;
+    return 0;
 }
 
 static long getAttrValue(DBADDR *paddr, short dbrType,
@@ -895,22 +921,23 @@ long dbGet(DBADDR *paddr, short dbrType,
     if (nRequest && *nRequest == 0)
         return 0;
 
-    if (!pfl || pfl->type == dbfl_type_rec) {
+    if (!pfl) {
         field_type = paddr->field_type;
         no_elements = capacity = paddr->no_elements;
-
-        /* Update field info from record
-         * may modify paddr->pfield
-         */
-        if (paddr->pfldDes->special == SPC_DBADDR &&
-            (prset = dbGetRset(paddr)) &&
-            prset->get_array_info) {
-            status = prset->get_array_info(paddr, &no_elements, &offset);
-        } else
-            offset = 0;
     } else {
         field_type = pfl->field_type;
         no_elements = capacity = pfl->no_elements;
+    }
+
+    /* Update field info from record (if necessary);
+     * may modify paddr->pfield.
+     */
+    if (!dbfl_has_copy(pfl) &&
+        paddr->pfldDes->special == SPC_DBADDR &&
+        (prset = dbGetRset(paddr)) &&
+        prset->get_array_info) {
+        status = prset->get_array_info(paddr, &no_elements, &offset);
+    } else {
         offset = 0;
     }
 
@@ -937,19 +964,27 @@ long dbGet(DBADDR *paddr, short dbrType,
     if (offset == 0 && (!nRequest || no_elements == 1)) {
         if (nRequest)
             *nRequest = 1;
-        if (!pfl || pfl->type == dbfl_type_rec) {
+        else if (no_elements < 1) {
+            status = S_db_onlyOne;
+            goto done;
+        }
+
+        if (!dbfl_has_copy(pfl)) {
             status = dbFastGetConvertRoutine[field_type][dbrType]
                 (paddr->pfield, pbuf, paddr);
         } else {
             DBADDR localAddr = *paddr; /* Structure copy */
 
+            if (no_elements < 1) {
+                status = S_db_badField;
+                goto done;
+            }
+
             localAddr.field_type = pfl->field_type;
             localAddr.field_size = pfl->field_size;
+            /* not used by dbFastConvert: */
             localAddr.no_elements = pfl->no_elements;
-            if (pfl->type == dbfl_type_val)
-                localAddr.pfield = (char *) &pfl->u.v.field;
-            else
-                localAddr.pfield = (char *)  pfl->u.r.field;
+            localAddr.pfield = dbfl_pfield(pfl);
             status = dbFastGetConvertRoutine[field_type][dbrType]
                 (localAddr.pfield, pbuf, &localAddr);
         }
@@ -960,6 +995,8 @@ long dbGet(DBADDR *paddr, short dbrType,
         if (nRequest) {
             if (no_elements < *nRequest)
                 *nRequest = no_elements;
+            if (capacity < *nRequest)
+                *nRequest = capacity;
             n = *nRequest;
         } else {
             n = 1;
@@ -976,19 +1013,22 @@ long dbGet(DBADDR *paddr, short dbrType,
         }
         /* convert data into the caller's buffer */
         if (n <= 0) {
-            ;/*do nothing*/
-        } else if (!pfl || pfl->type == dbfl_type_rec) {
+            ;                           /*do nothing */
+        } else if (!dbfl_has_copy(pfl)) {
             status = convert(paddr, pbuf, n, capacity, offset);
         } else {
             DBADDR localAddr = *paddr; /* Structure copy */
 
+            if (pfl->no_elements < 1) {
+                status = S_db_badField;
+                goto done;
+            }
+
             localAddr.field_type = pfl->field_type;
             localAddr.field_size = pfl->field_size;
+            /* not used by dbConvert, it uses the passed capacity instead: */
             localAddr.no_elements = pfl->no_elements;
-            if (pfl->type == dbfl_type_val)
-                localAddr.pfield = (char *) &pfl->u.v.field;
-            else
-                localAddr.pfield = (char *)  pfl->u.r.field;
+            localAddr.pfield = dbfl_pfield(pfl);
             status = convert(&localAddr, pbuf, n, capacity, offset);
         }
 
@@ -1010,7 +1050,7 @@ devSup* dbDTYPtoDevSup(dbRecordType *prdes, int dtyp) {
     return (devSup *)ellNth(&prdes->devList, dtyp+1);
 }
 
-devSup* dbDSETtoDevSup(dbRecordType *prdes, struct dset *pdset) {
+devSup* dbDSETtoDevSup(dbRecordType *prdes, dset *pdset) {
     devSup *pdevSup = (devSup *)ellFirst(&prdes->devList);
     while (pdevSup) {
         if (pdset == pdevSup->pdset) return pdevSup;
@@ -1023,7 +1063,7 @@ static long dbPutFieldLink(DBADDR *paddr,
     short dbrType, const void *pbuffer, long nRequest)
 {
     dbLinkInfo  link_info;
-    DBADDR      *pdbaddr = NULL;
+    dbChannel   *chan = NULL;
     dbCommon    *precord = paddr->precord;
     dbCommon    *lockrecs[2];
     dbLocker    locker;
@@ -1032,7 +1072,7 @@ static long dbPutFieldLink(DBADDR *paddr,
     struct link *plink = (struct link *)paddr->pfield;
     const char  *pstring = (const char *)pbuffer;
     struct dsxt *old_dsxt = NULL;
-    struct dset *new_dset = NULL;
+    dset *new_dset = NULL;
     struct dsxt *new_dsxt = NULL;
     devSup      *new_devsup = NULL;
     long        status;
@@ -1061,16 +1101,11 @@ static long dbPutFieldLink(DBADDR *paddr,
 
     if (link_info.ltype == PV_LINK &&
         (link_info.modifiers & (pvlOptCA | pvlOptCP | pvlOptCPP)) == 0) {
-        DBADDR tempaddr;
-
-        if (dbNameToAddr(link_info.target, &tempaddr)==0) {
-            /* This will become a DB link. */
-            pdbaddr = malloc(sizeof(*pdbaddr));
-            if (!pdbaddr) {
-                status = S_db_noMemory;
-                goto cleanup;
-            }
-            *pdbaddr = tempaddr; /* struct copy */
+        chan = dbChannelCreate(link_info.target);
+        if (chan && (status = dbChannelOpen(chan)) != 0) {
+            errlogPrintf(ERL_ERROR ": dbPutFieldLink %s.%s=%s: dbChannelOpen() failed w/ 0x%lx\n",
+                precord->name, pfldDes->name, link_info.target, status);
+            goto cleanup;
         }
     }
 
@@ -1079,7 +1114,7 @@ static long dbPutFieldLink(DBADDR *paddr,
 
     memset(&locker, 0, sizeof(locker));
     lockrecs[0] = precord;
-    lockrecs[1] = pdbaddr ? pdbaddr->precord : NULL;
+    lockrecs[1] = chan ? dbChannelRecord(chan) : NULL;
     dbLockerPrepare(&locker, lockrecs, 2);
 
     dbScanLockMany(&locker);
@@ -1167,7 +1202,8 @@ static long dbPutFieldLink(DBADDR *paddr,
     case PV_LINK:
     case CONSTANT:
     case JSON_LINK:
-        dbAddLink(&locker, plink, pfldDes->field_type, pdbaddr);
+        dbAddLink(&locker, plink, pfldDes->field_type, chan);
+        chan = NULL; /* don't clean it up */
         break;
 
     case DB_LINK:
@@ -1197,6 +1233,8 @@ unlock:
     dbScanUnlockMany(&locker);
     dbLockerFinalize(&locker);
 cleanup:
+    if (chan)
+        dbChannelDelete(chan);
     free(link_info.target);
     return status;
 }
@@ -1204,11 +1242,11 @@ cleanup:
 long dbPutField(DBADDR *paddr, short dbrType,
     const void *pbuffer, long nRequest)
 {
-    long	status = 0;
-    long	special  = paddr->special;
-    dbFldDes	*pfldDes = paddr->pfldDes;
-    dbCommon	*precord = paddr->precord;
-    short	dbfType  = paddr->field_type;
+    long        status = 0;
+    long        special  = paddr->special;
+    dbFldDes    *pfldDes = paddr->pfldDes;
+    dbCommon    *precord = paddr->precord;
+    short       dbfType  = paddr->field_type;
 
     if (special == SPC_ATTRIBUTE)
         return S_db_noMod;
@@ -1285,9 +1323,9 @@ long dbPut(DBADDR *paddr, short dbrType,
     void *pfieldsave  = paddr->pfield;
     rset *prset = dbGetRset(paddr);
     long status = 0;
-    long offset;
     dbFldDes *pfldDes;
     int isValueField;
+    int propertyUpdate = paddr->pfldDes->prop && precord->mlis.count;
 
     if (special == SPC_ATTRIBUTE)
         return S_db_noMod;
@@ -1309,38 +1347,59 @@ long dbPut(DBADDR *paddr, short dbrType,
         if (status) return status;
     }
 
-    if (paddr->pfldDes->special == SPC_DBADDR &&
-        prset && prset->get_array_info) {
-        long dummy;
+    if (nRequest>1 || paddr->pfldDes->special == SPC_DBADDR) {
+        long offset = 0;
+        if (paddr->pfldDes->special == SPC_DBADDR &&
+            prset && prset->get_array_info) {
+            long dummy;
 
-        status = prset->get_array_info(paddr, &dummy, &offset);
-        /* paddr->pfield may be modified */
-        if (status) goto done;
-    } else
-        offset = 0;
-
-    if (no_elements <= 1) {
-        status = dbFastPutConvertRoutine[dbrType][field_type](pbuffer,
-            paddr->pfield, paddr);
-        nRequest = 1;
-    } else {
+            status = prset->get_array_info(paddr, &dummy, &offset);
+            /* paddr->pfield may be modified */
+            if (status) goto done;
+        }
         if (no_elements < nRequest)
             nRequest = no_elements;
         status = dbPutConvertRoutine[dbrType][field_type](paddr, pbuffer,
             nRequest, no_elements, offset);
+        /* update array info */
+        if (!status && paddr->pfldDes->special == SPC_DBADDR &&
+                prset && prset->put_array_info) {
+            status = prset->put_array_info(paddr, nRequest);
+        }
+    } else {
+        if (nRequest < 1) {
+            recGblSetSevr(precord, LINK_ALARM, INVALID_ALARM);
+        } else {
+            if (propertyUpdate && paddr->field_size <= MAX_STRING_SIZE) {
+                char propBuffer[MAX_STRING_SIZE];
+                status = dbFastPutConvertRoutine[dbrType][field_type](pbuffer,
+                    &propBuffer, paddr);
+                if (!status) {
+                    if (memcmp(paddr->pfield, &propBuffer, paddr->field_size) != 0) {
+                        memcpy(paddr->pfield, &propBuffer, paddr->field_size);
+                    } else {
+                        /* suppress DBE_PROPERTY event if property did not change */
+                        propertyUpdate = 0;
+                    }
+                }
+            } else {
+                status = dbFastPutConvertRoutine[dbrType][field_type](pbuffer,
+                    paddr->pfield, paddr);
+            }
+            nRequest = 1;
+        }
     }
 
-    /* update array info */
-    if (!status &&
-        paddr->pfldDes->special == SPC_DBADDR &&
-        prset && prset->put_array_info) {
-        status = prset->put_array_info(paddr, nRequest);
-    }
+    /* Post property updates before second dbPutSpecial */
+    /* which may post DBE_VALUE and/or DBE_LOG events */
+    if (propertyUpdate && !status)
+        db_post_events(precord, NULL, DBE_PROPERTY);
 
     /* Always do special processing if needed */
     if (special) {
         long status2 = dbPutSpecial(paddr, 1);
-        if (status2) goto done;
+        if (status2)
+            status = status2;
     }
     if (status) goto done;
 
@@ -1352,12 +1411,6 @@ long dbPut(DBADDR *paddr, short dbrType,
     if (precord->mlis.count &&
         !(isValueField && pfldDes->process_passive))
         db_post_events(precord, pfieldsave, DBE_VALUE | DBE_LOG);
-    /* If this field is a property (metadata) field,
-     * then post a property change event (even if the field
-     * didn't change).
-     */
-    if (precord->mlis.count && pfldDes->prop)
-        db_post_events(precord, NULL, DBE_PROPERTY);
 done:
     paddr->pfield = pfieldsave;
     return status;
