@@ -3,6 +3,7 @@
 *     National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 *     Operator of Los Alamos National Laboratory.
+* SPDX-License-Identifier: EPICS
 * EPICS BASE is distributed subject to a Software License Agreement found
 * in file LICENSE that is included with this distribution.
 \*************************************************************************/
@@ -25,7 +26,6 @@
 #include "errlog.h"
 #include "errMdef.h"
 
-#define epicsExportSharedSymbols
 #include "dbAccessDefs.h"
 #include "dbAddr.h"
 #include "dbBase.h"
@@ -302,7 +302,7 @@ static long getStringUlong(const dbAddr *paddr,
                 epicsFloat64 dval;
 
                 status = epicsParseFloat64(psrc, &dval, &end);
-                if (!status && 0 <= dval && dval <= ULONG_MAX)
+                if (!status && 0 <= dval && dval <= UINT_MAX)
                     *pdst = dval;
             }
             if (status)
@@ -1052,7 +1052,7 @@ static long putStringUlong(dbAddr *paddr,
             epicsFloat64 dval;
 
             status = epicsParseFloat64(psrc, &dval, &end);
-            if (!status && 0 <= dval && dval <= ULONG_MAX)
+            if (!status && 0 <= dval && dval <= UINT_MAX)
                 *pdst = dval;
         }
         if (status)
@@ -1704,7 +1704,7 @@ static long putEnumEnum PUT_NOCONVERT(epicsEnum16, epicsEnum16)
  DBR_FLOAT,       DBR_DOUBLE,      DBR_ENUM
  ***************************************************************************/
 
-epicsShareDef GETCONVERTFUNC dbGetConvertRoutine[DBF_DEVICE+1][DBR_ENUM+1] = {
+GETCONVERTFUNC dbGetConvertRoutine[DBF_DEVICE+1][DBR_ENUM+1] = {
 
 /* source is a DBF_STRING               */
 {getStringString, getStringChar,   getStringUchar,  getStringShort,  getStringUshort,
@@ -1775,7 +1775,7 @@ epicsShareDef GETCONVERTFUNC dbGetConvertRoutine[DBF_DEVICE+1][DBR_ENUM+1] = {
  DBF_MENU,        DBF_DEVICE
  ***************************************************************************/
 
-epicsShareDef PUTCONVERTFUNC dbPutConvertRoutine[DBR_ENUM+1][DBF_DEVICE+1] = {
+PUTCONVERTFUNC dbPutConvertRoutine[DBR_ENUM+1][DBF_DEVICE+1] = {
 /* source is a DBR_STRING               */
 {putStringString, putStringChar,   putStringUchar,  putStringShort,  putStringUshort,
  putStringLong,   putStringUlong,  putStringInt64,  putStringUInt64,
