@@ -5,6 +5,7 @@
 *     Operator of Los Alamos National Laboratory.
 * Copyright (c) 2013 Helmholtz-Zentrum Berlin
 *     für Materialien und Energie GmbH.
+* SPDX-License-Identifier: EPICS
 * EPICS BASE is distributed subject to a Software License Agreement found
 * in file LICENSE that is included with this distribution.
 \*************************************************************************/
@@ -36,7 +37,6 @@
 #include "epicsTime.h"
 #include "taskwd.h"
 
-#define epicsExportSharedSymbols
 #include "callback.h"
 #include "dbAccessDefs.h"
 #include "dbAddr.h"
@@ -155,7 +155,7 @@ void scanStop(void)
 {
     int i;
 
-    if (scanCtl == ctlExit) return;
+    if (scanCtl == ctlInit || scanCtl == ctlExit) return;
     scanCtl = ctlExit;
 
     interruptAccept = FALSE;
@@ -168,9 +168,13 @@ void scanStop(void)
         epicsEventSignal(ppsl->loopEvent);
         epicsEventWait(startStopEvent);
     }
+    for (i = 0; i < nPeriodic; i++) {
+        epicsThreadMustJoin(periodicTaskId[i]);
+    }
 
     scanOnce((dbCommon *)&exitOnce);
     epicsEventWait(startStopEvent);
+    epicsThreadMustJoin(onceTaskId);
 }
 
 void scanCleanup(void)
@@ -260,7 +264,7 @@ void scanAdd(struct dbCommon *precord)
     } else if (scan == menuScanI_O_Intr) {
         ioscan_head *piosh = NULL;
         int prio;
-        DEVSUPFUN get_ioint_info;
+        long (*get_ioint_info)(int, struct dbCommon *, IOSCANPVT*);
 
         if (precord->dset == NULL){
             recGblRecordError(-1, (void *)precord,
@@ -328,7 +332,7 @@ void scanDelete(struct dbCommon *precord)
     } else if (scan == menuScanI_O_Intr) {
         ioscan_head *piosh = NULL;
         int prio;
-        DEVSUPFUN get_ioint_info;
+        long (*get_ioint_info)(int, struct dbCommon *, IOSCANPVT*);
 
         if (precord->dset==NULL) {
             recGblRecordError(-1, (void *)precord,
@@ -761,14 +765,16 @@ void scanOnceQueueShow(const int reset)
 
 static void initOnce(void)
 {
+    epicsThreadOpts opts = EPICS_THREAD_OPTS_INIT;
+    opts.joinable = 1;
+    opts.priority = epicsThreadPriorityScanLow + nPeriodic;
+    opts.stackSize = epicsThreadStackBig;
     if ((onceQ = epicsRingBytesLockedCreate(sizeof(onceEntry)*onceQueueSize)) == NULL) {
         cantProceed("initOnce: Ring buffer create failed\n");
     }
     if(!onceSem)
         onceSem = epicsEventMustCreate(epicsEventEmpty);
-    onceTaskId = epicsThreadCreate("scanOnce",
-        epicsThreadPriorityScanLow + nPeriodic,
-        epicsThreadGetStackSize(epicsThreadStackBig), onceTask, 0);
+    onceTaskId = epicsThreadCreateOpt("scanOnce", onceTask, 0, &opts);
 
     epicsEventWait(startStopEvent);
 }
@@ -817,7 +823,7 @@ static void periodicTask(void *arg)
             epicsTimeAddSeconds(&next, delay);
             if (++overruns >= 10 &&
                 epicsTimeDiffInSeconds(&now, &reported) > report_delay) {
-                errlogPrintf("\ndbScan warning from '%s' scan thread:\n"
+                errlogPrintf("\ndbScan " ERL_WARNING " from '%s' scan thread:\n"
                     "\tScan processing averages %.3f seconds (%.3f .. %.3f).\n"
                     "\tOver-runs have now happened %u times in a row.\n"
                     "\tTo fix this, move some records to a slower scan rate.\n",
@@ -932,14 +938,16 @@ static void spawnPeriodic(int ind)
 {
     periodic_scan_list *ppsl = papPeriodic[ind];
     char taskName[20];
+    epicsThreadOpts opts = EPICS_THREAD_OPTS_INIT;
+    opts.joinable = 1;
+    opts.priority = epicsThreadPriorityScanLow + ind;
+    opts.stackSize = epicsThreadStackBig;
 
     if (!ppsl) return;
 
     sprintf(taskName, "scan-%g", ppsl->period);
-    periodicTaskId[ind] = epicsThreadCreate(
-        taskName, epicsThreadPriorityScanLow + ind,
-        epicsThreadGetStackSize(epicsThreadStackBig),
-        periodicTask, (void *)ppsl);
+    periodicTaskId[ind] = epicsThreadCreateOpt(
+        taskName, periodicTask, (void *)ppsl, &opts);
 
     epicsEventWait(startStopEvent);
 }
@@ -1075,13 +1083,10 @@ static void addToList(struct dbCommon *precord, scan_list *psl)
     pse->pscan_list = psl;
     ptemp = (scan_element *)ellLast(&psl->list);
     while (ptemp) {
-        if (ptemp->precord->phas <= precord->phas) {
-            ellInsert(&psl->list, &ptemp->node, &pse->node);
-            break;
-        }
+        if (ptemp->precord->phas <= precord->phas) break;
         ptemp = (scan_element *)ellPrevious(&ptemp->node);
     }
-    if (ptemp == NULL) ellAdd(&psl->list, (void *)pse);
+    ellInsert(&psl->list, (ptemp ? &ptemp->node : NULL), &pse->node);
     psl->modified = TRUE;
     epicsMutexUnlock(psl->lock);
 }
@@ -1107,7 +1112,7 @@ static void deleteFromList(struct dbCommon *precord, scan_list *psl)
         return;
     }
     pse->pscan_list = NULL;
-    ellDelete(&psl->list, (void *)pse);
+    ellDelete(&psl->list, &pse->node);
     psl->modified = TRUE;
     epicsMutexUnlock(psl->lock);
 }
