@@ -3,6 +3,7 @@
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
+* SPDX-License-Identifier: EPICS
 * EPICS Base is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution.
 \*************************************************************************/
@@ -26,6 +27,10 @@
 #include <osiFileName.h>
 #include <osiUnistd.h>
 
+#ifdef _WIN32
+#include <shlwapi.h>
+#endif
+
 #define MAX_BUFFER_SIZE 4096
 #define MAX_DEPS 1024
 
@@ -60,7 +65,7 @@ typedef struct inputData inputData;
 static void inputConstruct(inputData **ppvt);
 static void inputDestruct(inputData * const pvt);
 static void inputAddPath(inputData * const pvt, const char * const pval);
-static void inputBegin(inputData * const pvt, const char * const fileName);
+static void inputBegin(inputData * const pvt, const char * const fileName, bool fromCmdLine);
 static char *inputNextLine(inputData * const pvt);
 static void inputNewIncludeFile(inputData * const pvt, const char * const name);
 static void inputErrPrint(const inputData * const pvt);
@@ -82,7 +87,8 @@ static void addMacroReplacements(MAC_HANDLE * const macPvt,
                                  const char * const pval);
 static void makeSubstitutions(inputData * const inputPvt,
                               MAC_HANDLE * const macPvt,
-                              const char * const templateName);
+                              const char * const templateName,
+                              bool fromCmdLine);
 
 /*Global variables */
 static int opt_V = 0;
@@ -170,7 +176,7 @@ int main(int argc,char **argv)
 
     if (substitutionName.empty()) {
         STEP("Single template+substitutions file");
-        makeSubstitutions(inputPvt, macPvt, templateName);
+        makeSubstitutions(inputPvt, macPvt, templateName, true);
     }
     else {
         subInfo *substitutePvt;
@@ -202,7 +208,7 @@ int main(int argc,char **argv)
                         macPushScope(macPvt);
 
                     addMacroReplacements(macPvt, macStr);
-                    makeSubstitutions(inputPvt, macPvt, filename);
+                    makeSubstitutions(inputPvt, macPvt, filename, false);
 
                     if (localScope)
                         macPopScope(macPvt);
@@ -263,7 +269,7 @@ static void addMacroReplacements(MAC_HANDLE * const macPvt,
     if (status) {
         status = macInstallMacros(macPvt, pairs);
         if (!status) {
-            fprintf(stderr, "Error from macInstallMacros\n");
+            fprintf(stderr, ERL_ERROR " from macInstallMacros\n");
             usageExit(1);
         }
         free(pairs);
@@ -275,14 +281,15 @@ static const char *cmdNames[] = {"include","substitute"};
 
 static void makeSubstitutions(inputData * const inputPvt,
                               MAC_HANDLE * const macPvt,
-                              const char * const templateName)
+                              const char * const templateName,
+                              bool fromCmdLine)
 {
     char *input;
     static char buffer[MAX_BUFFER_SIZE];
     int  n;
 
     ENTER;
-    inputBegin(inputPvt, templateName);
+    inputBegin(inputPvt, templateName, fromCmdLine);
     while ((input = inputNextLine(inputPvt))) {
         int     expand=1;
         char    *p;
@@ -300,11 +307,11 @@ static void makeSubstitutions(inputData * const inputPvt,
             char *pstart;
             char *pend;
             int  cmdind=-1;
-            int  i;
+            size_t  i;
 
             for (i = 0; i < NELEMENTS(cmdNames); i++) {
                 if (strstr(command, cmdNames[i])) {
-                    cmdind = i;
+                    cmdind = (int)i;
                 }
             }
             if (cmdind < 0) goto endcmd;
@@ -316,7 +323,7 @@ static void makeSubstitutions(inputData * const inputPvt,
             pstart = ++p;
             /*Look for end quote*/
             while (*p && (*p != '"')) {
-                /*allow escape for embeded quote*/
+                /*allow escape for embedded quote*/
                 if ((p[0] == '\\') && p[1] == '"') {
                     p += 2;
                     continue;
@@ -377,7 +384,7 @@ struct inputData {
     inputData() { memset(inputBuffer, 0, sizeof(inputBuffer) * sizeof(inputBuffer[0])); };
 };
 
-static void inputOpenFile(inputData *pinputData, const char * const filename);
+static void inputOpenFile(inputData *pinputData, const char * const filename, bool fromCmdLine);
 static void inputCloseFile(inputData *pinputData);
 static void inputCloseAllFiles(inputData *pinputData);
 
@@ -430,11 +437,11 @@ static void inputAddPath(inputData * const pinputData, const char * const path)
     EXIT;
 }
 
-static void inputBegin(inputData * const pinputData, const char * const fileName)
+static void inputBegin(inputData * const pinputData, const char * const fileName, bool fromCmdLine)
 {
     ENTER;
     inputCloseAllFiles(pinputData);
-    inputOpenFile(pinputData, fileName);
+    inputOpenFile(pinputData, fileName, fromCmdLine);
     EXIT;
 }
 
@@ -461,7 +468,7 @@ static void inputNewIncludeFile(inputData * const pinputData,
                                 const char * const name)
 {
     ENTER;
-    inputOpenFile(pinputData,name);
+    inputOpenFile(pinputData, name, false);
     EXIT;
 }
 
@@ -492,7 +499,15 @@ static void inputErrPrint(const inputData *const pinputData)
     EXIT;
 }
 
-static void inputOpenFile(inputData *pinputData, const char * const filename)
+static int isPathRelative(const char * const path) {
+#ifdef _WIN32
+    return path && PathIsRelativeA(path);
+#else
+    return path && path[0] != '/';
+#endif
+}
+
+static void inputOpenFile(inputData *pinputData, const char * const filename, bool fromCmdLine)
 {
     std::list<std::string>& pathList = pinputData->pathList;
     std::list<std::string>::iterator pathIt = pathList.end();
@@ -504,7 +519,7 @@ static void inputOpenFile(inputData *pinputData, const char * const filename)
         STEP("Using stdin");
         fp = stdin;
     }
-    else if (pathList.empty() || strchr(filename, '/')){
+    else if (fromCmdLine || pathList.empty() || !isPathRelative(filename)){
         STEPS("Opening ", filename);
         fp = fopen(filename, "r");
     }
@@ -521,7 +536,7 @@ static void inputOpenFile(inputData *pinputData, const char * const filename)
     }
 
     if (!fp) {
-        fprintf(stderr, "msi: Can't open file '%s'\n", filename);
+        fprintf(stderr, ERL_ERROR " msi: Can't open file '%s'\n", filename);
         inputErrPrint(pinputData);
         abortExit(1);
     }
@@ -671,7 +686,7 @@ static void substituteOpen(subInfo **ppvt, const std::string& substitutionName)
 
     fp = fopen(substitutionName.c_str(), "r");
     if (!fp) {
-        fprintf(stderr, "msi: Can't open file '%s'\n", substitutionName.c_str());
+        fprintf(stderr, ERL_ERROR " msi: Can't open file '%s'\n", substitutionName.c_str());
         abortExit(1);
     }
 
@@ -1027,7 +1042,7 @@ static tokenType subGetNextToken(subFile *psubFile)
                 subFileErrPrint(psubFile, "Strings must be on single line\n");
                 abortExit(1);
             }
-            /*allow  escape for embeded quote*/
+            /*allow  escape for embedded quote*/
             if ((p[0] == '\\') && p[1] == '"') {
                 *pto++ = *p++;
                 *pto++ = *p++;
