@@ -5,6 +5,7 @@
 *     Operator of Los Alamos National Laboratory.
 * Copyright (c) 2013 Helmholtz-Zentrum Berlin
 *     für Materialien und Energie GmbH.
+* SPDX-License-Identifier: EPICS
 * EPICS BASE is distributed subject to a Software License Agreement found
 * in file LICENSE that is included with this distribution.
 \*************************************************************************/
@@ -70,9 +71,7 @@
 #include "registryJLinks.h"
 #include "registryRecordType.h"
 
-static enum {
-    iocVirgin, iocBuilding, iocBuilt, iocRunning, iocPaused, iocStopped
-} iocState = iocVirgin;
+static enum iocStateEnum iocState = iocVoid;
 static enum {
     buildServers, buildIsolated
 } iocBuildMode;
@@ -99,6 +98,11 @@ static void iterateRecords(recIterFunc func, void *user);
 int dbThreadRealtimeLock = 1;
 epicsExportAddress(int, dbThreadRealtimeLock);
 
+enum iocStateEnum getIocState(void)
+{
+    return iocState;
+}
+
 /*
  *  Initialize EPICS on the IOC.
  */
@@ -109,8 +113,8 @@ int iocInit(void)
 
 static int iocBuild_1(void)
 {
-    if (iocState != iocVirgin && iocState != iocStopped) {
-        errlogPrintf("iocBuild: IOC can only be initialized from uninitialized or stopped state\n");
+    if (iocState != iocVoid) {
+        errlogPrintf("iocBuild: " ERL_ERROR " IOC can only be initialized from uninitialized or stopped state\n");
         return -1;
     }
     errlogInit(0);
@@ -122,7 +126,7 @@ static int iocBuild_1(void)
 
     errlogPrintf("Starting iocInit\n");
     if (checkDatabase(pdbbase)) {
-        errlogPrintf("iocBuild: Aborting, bad database definition (DBD)!\n");
+        errlogPrintf("iocBuild: " ERL_ERROR " Aborting, bad database definition (DBD)!\n");
         return -1;
     }
     epicsSignalInstallSigHupIgnore();
@@ -169,7 +173,7 @@ static int iocBuild_2(void)
 
     scanInit();
     if (asInit()) {
-        errlogPrintf("iocBuild: asInit Failed.\n");
+        errlogPrintf(ERL_ERROR " iocBuild: asInit Failed.\n");
         return -1;
     }
     dbProcessNotifyInit();
@@ -233,7 +237,7 @@ int iocBuildIsolated(void)
 int iocRun(void)
 {
     if (iocState != iocPaused && iocState != iocBuilt) {
-        errlogPrintf("iocRun: IOC not paused\n");
+        errlogPrintf("iocRun: " ERL_WARNING " IOC not paused\n");
         return -1;
     }
     initHookAnnounce(initHookAtIocRun);
@@ -264,7 +268,7 @@ int iocRun(void)
 int iocPause(void)
 {
     if (iocState != iocRunning) {
-        errlogPrintf("iocPause: IOC not running\n");
+        errlogPrintf("iocPause: " ERL_WARNING " IOC not running\n");
         return -1;
     }
     initHookAnnounce(initHookAtIocPause);
@@ -302,59 +306,59 @@ static int checkDatabase(dbBase *pdbbase)
     const dbMenu *pMenu;
 
     if (!pdbbase) {
-        errlogPrintf("checkDatabase: No database definitions loaded.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " No database definitions loaded.\n");
         return -1;
     }
 
     pMenu = dbFindMenu(pdbbase, "menuConvert");
     if (!pMenu) {
-        errlogPrintf("checkDatabase: menuConvert not defined.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuConvert not defined.\n");
         return -1;
     }
     if (pMenu->nChoice <= menuConvertLINEAR) {
-        errlogPrintf("checkDatabase: menuConvert has too few choices.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuConvert has too few choices.\n");
         return -1;
     }
     if (strcmp(pMenu->papChoiceName[menuConvertNO_CONVERSION],
                "menuConvertNO_CONVERSION")) {
-        errlogPrintf("checkDatabase: menuConvertNO_CONVERSION doesn't match.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuConvertNO_CONVERSION doesn't match.\n");
         return -1;
     }
     if (strcmp(pMenu->papChoiceName[menuConvertSLOPE], "menuConvertSLOPE")) {
-        errlogPrintf("checkDatabase: menuConvertSLOPE doesn't match.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuConvertSLOPE doesn't match.\n");
         return -1;
     }
     if (strcmp(pMenu->papChoiceName[menuConvertLINEAR], "menuConvertLINEAR")) {
-        errlogPrintf("checkDatabase: menuConvertLINEAR doesn't match.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuConvertLINEAR doesn't match.\n");
         return -1;
     }
 
     pMenu = dbFindMenu(pdbbase, "menuScan");
     if (!pMenu) {
-        errlogPrintf("checkDatabase: menuScan not defined.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuScan not defined.\n");
         return -1;
     }
     if (pMenu->nChoice <= menuScanI_O_Intr) {
-        errlogPrintf("checkDatabase: menuScan has too few choices.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuScan has too few choices.\n");
         return -1;
     }
     if (strcmp(pMenu->papChoiceName[menuScanPassive],
                "menuScanPassive")) {
-        errlogPrintf("checkDatabase: menuScanPassive doesn't match.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuScanPassive doesn't match.\n");
         return -1;
     }
     if (strcmp(pMenu->papChoiceName[menuScanEvent],
                "menuScanEvent")) {
-        errlogPrintf("checkDatabase: menuScanEvent doesn't match.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuScanEvent doesn't match.\n");
         return -1;
     }
     if (strcmp(pMenu->papChoiceName[menuScanI_O_Intr],
                "menuScanI_O_Intr")) {
-        errlogPrintf("checkDatabase: menuScanI_O_Intr doesn't match.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuScanI_O_Intr doesn't match.\n");
         return -1;
     }
     if (pMenu->nChoice <= SCAN_1ST_PERIODIC) {
-        errlogPrintf("checkDatabase: menuScan has no periodic choices.\n");
+        errlogPrintf("checkDatabase: " ERL_ERROR " menuScan has no periodic choices.\n");
         return -1;
     }
 
@@ -384,7 +388,7 @@ static void initDrvSup(void) /* Locate all driver support entry tables */
 
     for (pdrvSup = (drvSup *)ellFirst(&pdbbase->drvList); pdrvSup;
          pdrvSup = (drvSup *)ellNext(&pdrvSup->node)) {
-        struct drvet *pdrvet = registryDriverSupportFind(pdrvSup->name);
+        drvet *pdrvet = registryDriverSupportFind(pdrvSup->name);
 
         if (!pdrvet) {
             errlogPrintf("iocInit: driver %s not found\n", pdrvSup->name);
@@ -433,7 +437,7 @@ static void initDevSup(void)
         for (pdevSup = (devSup *)ellFirst(&pdbRecordType->devList);
              pdevSup;
              pdevSup = (devSup *)ellNext(&pdevSup->node)) {
-            struct dset *pdset = registryDeviceSupportFind(pdevSup->name);
+            dset *pdset = registryDeviceSupportFind(pdevSup->name);
 
             if (!pdset) {
                 errlogPrintf("device support %s not found\n",pdevSup->name);
@@ -456,7 +460,7 @@ static void finishDevSup(void)
         for (pdevSup = (devSup *)ellFirst(&pdbRecordType->devList);
              pdevSup;
              pdevSup = (devSup *)ellNext(&pdevSup->node)) {
-            struct dset *pdset = pdevSup->pdset;
+            dset *pdset = pdevSup->pdset;
 
             if (pdset && pdset->init)
                 pdset->init(1);
@@ -505,7 +509,7 @@ static void doInitRecord0(dbRecordType *pdbRecordType, dbCommon *precord,
 
     /* Initial UDF severity */
     if (precord->udf && precord->stat == UDF_ALARM)
-    	precord->sevr = precord->udfs;
+        precord->sevr = precord->udfs;
 
     /* Init DSET NOTE that result may be NULL */
     pdevSup = dbDTYPtoDevSup(pdbRecordType, precord->dtyp);
@@ -705,23 +709,29 @@ static void doFreeRecord(dbRecordType *pdbRecordType, dbCommon *precord,
 
 int iocShutdown(void)
 {
-    if (iocState == iocVirgin || iocState == iocStopped)
-        return 0;
+    if (iocState == iocVoid) return 0;
+
+    initHookAnnounce(initHookAtShutdown);
 
     iterateRecords(doCloseLinks, NULL);
+    initHookAnnounce(initHookAfterCloseLinks);
 
-    if (iocBuildMode == buildIsolated) {
-        /* stop and "join" threads */
-        scanStop();
-        callbackStop();
-    }
-    else
+    /* stop and "join" threads */
+    scanStop();
+    initHookAnnounce(initHookAfterStopScan);
+    callbackStop();
+    initHookAnnounce(initHookAfterStopCallback);
+
+    if (iocBuildMode != buildIsolated) {
         dbStopServers();
+    }
 
     dbCaShutdown(); /* must be before dbFreeRecord and dbChannelExit */
+    initHookAnnounce(initHookAfterStopLinks);
 
     if (iocBuildMode == buildIsolated) {
         /* free resources */
+        initHookAnnounce(initHookBeforeFree);
         scanCleanup();
         callbackCleanup();
 
@@ -734,8 +744,10 @@ int iocShutdown(void)
         iocshFree();
     }
 
-    iocState = iocStopped;
+    iocState = iocVoid;
     iocBuildMode = buildServers;
+
+    initHookAnnounce(initHookAfterShutdown);
     return 0;
 }
 
