@@ -1,3 +1,7 @@
+/* SPDX-FileCopyrightText: 2015 Argonne National Laboratory */
+
+/* SPDX-License-Identifier: EPICS */
+
 /* dev_APPNAME_Version.c */
 /* Example device support for the lsi (long string input) record
  * providing the module version string as the value
@@ -33,6 +37,6 @@ static long read_string(lsiRecord *prec)
 }
 
 static lsidset dev_CSAFEAPPNAME_Version = {
-    5, NULL, NULL, NULL, NULL, read_string
+	{5, NULL, NULL, NULL, NULL}, read_string
 };
 epicsExportAddress(dset,dev_CSAFEAPPNAME_Version);
