@@ -1,3 +1,7 @@
+/* SPDX-FileCopyrightText: 2008 Argonne National Laboratory */
+
+/* SPDX-License-Identifier: EPICS */
+
 /* initTrace.c */
 
 /*
@@ -28,7 +32,14 @@ int traceIocInit(void) {
 }
 
 
-static const iocshFuncDef traceInitFuncDef = {"traceIocInit", 0, NULL};
+static const iocshFuncDef traceInitFuncDef = {
+    "traceIocInit",
+    0,
+    NULL,
+    "Setup the 'iocInit' function to trace its different stages.\n"
+    "This function must be called before 'iocInit'.\n\n"
+    "Example: traceIocInit\n",
+};
 static void traceInitFunc(const iocshArgBuf *args) {
     traceIocInit();
 }
