@@ -29,7 +29,7 @@ using std::endl;
 using std::vector;
 using namespace epics::pvData;
 
-namespace epics { namespace pvCopy { 
+namespace epics { namespace pvCopy {
 
 /**
  * Convenience method for implementing dump.
@@ -60,19 +60,19 @@ struct CopyNode {
     PVStructurePtr options;
     vector<PVFilterPtr> pvFilters;
 };
-    
+
 static CopyNodePtr NULLCopyNode;
 
 typedef std::vector<CopyNodePtr> CopyNodePtrArray;
 typedef std::tr1::shared_ptr<CopyNodePtrArray> CopyNodePtrArrayPtr;
-    
+
 struct CopyStructureNode : public  CopyNode {
     CopyNodePtrArrayPtr nodes;
 };
 
 PVCopyPtr PVCopy::create(
-    PVStructurePtr const &pvMaster, 
-    PVStructurePtr const &pvRequest, 
+    PVStructurePtr const &pvMaster,
+    PVStructurePtr const &pvRequest,
     string const & structureName)
 {
     PVStructurePtr pvStructure(pvRequest);
@@ -88,7 +88,6 @@ PVCopyPtr PVCopy::create(
     bool result = pvCopy->init(pvStructure);
     if(!result) return PVCopyPtr();
     pvCopy->traverseMasterInitPlugin();
-//cout << pvCopy->dump() << endl;
     return pvCopy;
 }
 
@@ -114,7 +113,7 @@ PVStructurePtr PVCopy::createPVStructure()
         cacheInitStructure.reset();
         return save;
     }
-    PVStructurePtr pvStructure = 
+    PVStructurePtr pvStructure =
         getPVDataCreate()->createPVStructure(structure);
     return pvStructure;
 }
@@ -244,7 +243,7 @@ void PVCopy::updateMasterCheckBitSet(
         bitSet->clear(nextSet);
         PVStructurePtr pv = static_pointer_cast<PVStructure>(pvField);
         PVFieldPtrArray pvFieldArray = pv->getPVFields();
-        for(size_t i=0; i>pvFieldArray.size(); ++i) {
+        for(size_t i=0; i<pvFieldArray.size(); ++i) {
             PVFieldPtr pvField = pvFieldArray[i];
             bitSet->set(pvField->getFieldOffset());
         }
@@ -296,7 +295,7 @@ void PVCopy::updateMaster(
         bitSet->clear(nextSet);
     }
 }
- 
+
 PVStructurePtr PVCopy::getOptions(std::size_t fieldOffset)
 {
     if(fieldOffset==0) return headNode->options;
@@ -414,7 +413,7 @@ void PVCopy::updateCopyFromBitSet(
     size_t offset = structureNode->structureOffset;
     size_t nextSet = bitSet->nextSetBit(offset);
     if(nextSet==string::npos) return;
-    if(offset>=pvCopy->getNextFieldOffset()) return;  
+    if(offset>=pvCopy->getNextFieldOffset()) return;
     PVStructurePtr pvCopyStructure = static_pointer_cast<PVStructure>(pvCopy);
     PVFieldPtrArray const & pvCopyFields = pvCopyStructure->getPVFields();
     for(size_t i=0; i<pvCopyFields.size(); ++i) {
@@ -428,20 +427,28 @@ PVCopy::PVCopy(
 {
 }
 
-void PVCopy::destroy()
-{
-    headNode.reset();
-}
-
 bool PVCopy::init(epics::pvData::PVStructurePtr const &pvRequest)
 {
     PVStructurePtr pvMasterStructure = pvMaster;
     size_t len = pvRequest->getPVFields().size();
     bool entireMaster = false;
-    if(len==0) entireMaster = true;
+    requestHasMasterField = false;
     PVStructurePtr pvOptions;
-    if(len==1) {
-        pvOptions = pvRequest->getSubField<PVStructure>("_options");
+    if(len==0) {
+        entireMaster = true;
+    }
+    else {
+        // If "_" is in the request, but not in the master structure,
+        // then assume the top level PV structure is requested
+        PVStructurePtr masterFieldPtr = pvMaster->getSubField<PVStructure>("_");
+        PVStructurePtr requestFieldPtr = pvRequest->getSubField<PVStructure>("_");
+        if (requestFieldPtr) {
+            requestHasMasterField = true;
+        }
+        if (!masterFieldPtr && requestFieldPtr) {
+            entireMaster = true;
+            pvOptions = requestFieldPtr->getSubField<PVStructure>("_options");
+        }
     }
     if(entireMaster) {
         structure = pvMasterStructure->getStructure();
@@ -529,9 +536,9 @@ CopyNodePtr PVCopy::createStructureNodes(
     for(size_t i=0; i<number; i++) {
         PVFieldPtr copyPVField = copyPVFields[i];
         string fieldName = copyPVField->getFieldName();
-        PVStructurePtr requestPVStructure = 
+        PVStructurePtr requestPVStructure =
              pvFromRequest->getSubField<PVStructure>(fieldName);
-        PVStructurePtr pvSubFieldOptions = 
+        PVStructurePtr pvSubFieldOptions =
             requestPVStructure->getSubField<PVStructure>("_options");
         PVFieldPtr pvMasterField = pvMasterStructure->getSubField(fieldName);
         if(!pvMasterField) {
@@ -641,14 +648,14 @@ void PVCopy::initPlugin(
 
 void PVCopy::traverseMasterInitPlugin()
 {
-    traverseMasterInitPlugin(headNode); 
+    traverseMasterInitPlugin(headNode);
 }
 
 void PVCopy::traverseMasterInitPlugin(CopyNodePtr const & node)
 {
     PVFieldPtr pvField = node->masterPVField;
     PVStructurePtr pvOptions = node->options;
-    if(pvOptions) initPlugin(node,pvOptions,pvField);	
+    if(pvOptions) initPlugin(node,pvOptions,pvField);
     if(!node->isStructure) return;
     CopyStructureNodePtr structureNode = static_pointer_cast<CopyStructureNode>(node);
     CopyNodePtrArrayPtr nodes = structureNode->nodes;
@@ -667,7 +674,7 @@ CopyNodePtr PVCopy::getCopyOffset(
         CopyNodePtr node = (*nodes)[i];
         if(!node->isStructure) {
             size_t off = node->masterPVField->getFieldOffset();
-            size_t nextOffset = node->masterPVField->getNextFieldOffset(); 
+            size_t nextOffset = node->masterPVField->getNextFieldOffset();
             if(offset>= off && offset<nextOffset) return node;
         } else {
             CopyStructureNodePtr subNode =
@@ -710,7 +717,7 @@ void PVCopy::setIgnore(CopyNodePtr const &node) {
          CopyNodePtrArrayPtr nodes = structureNode->nodes;
          for(size_t i=0; i<nodes->size(); ++i) {
             CopyNodePtr node = (*nodes)[i];
-            setIgnore(node);		}
+            setIgnore(node);        }
     } else {
         size_t num = node->masterPVField->getNumberFields();
         if(num>1) {
